/****************************************************************************
* FENPRINT Vn 1.0  November 2004
* By Mike Eggleston (Fenland Software Ltd)
*
* Copyright (c) 2004 Fenland Software Ltd 
*****************************************************************************/
#define PROGNAME	"FENPRINT"
#define DESC		"Fenland Software's Virtual Printing Utility"
#define VERSION		"1.0"
#define COPYRIGHT	"Copyright (c) 2004 Fenland Software Ltd"

#include "fenprint.h"

char printer_list[_PATH_LEN + 1] = "";
char printer_path[_PATH_LEN + 1] = "";
char print_to[_PATH_LEN + 1] = "";

void usage(void)
{
  fprintf(stderr, "%s Vn %s\n%s\n", PROGNAME,VERSION,COPYRIGHT);
  fprintf(stderr, "%s\n", DESC);
  fprintf(stderr, "fenprint [-P printer] [-T title] [-J description] [filename ...]\n");
  fprintf(stderr, "Please see the manuals or 'info' page for operating instructions\n\n");
  exit(0);
}

/****************************************************************************
* Load the configuration
****************************************************************************/
int freadln(FILE *fp,char *str)
{
  if(fgets(str,_PATH_LEN,fp) != NULL)
    if(str[0] != '#')
      return 1;
  return 0;
}


int get_param(char *d_str, char *s_str, char *key, int max)
{
  char tok[_PATH_LEN + 1], param[_PATH_LEN + 1];

  tok[0] = 0;
  param[0] = 0;
  if(sscanf(s_str,"%s %s",tok,param) == 0)
    return 0;

  if(!strcmp(tok,key))
  {
    char *s;

    s = strstr(s_str + strlen(tok),param);
    if(s != NULL) strncpy(d_str,s,max);
    d_str[max] = 0;
    s = strchr(d_str,'\n');
    if(s != NULL) *s = 0;
    s = strchr(d_str,'\r');
    if(s != NULL) *s = 0;

    return 1;
  }

  return 0;
}

int load_conf(void)
{
  FILE *fp;
  char str[_PATH_LEN + 1];
  char substr[_PATH_LEN + 1];

  if((fp = fopen(_CONFIG_FILE, "r")) == NULL)
  {
    fprintf(stderr,"ERROR - Can't open configuration file - %s\n", _CONFIG_FILE);
    return(0);
  }

  while(!feof(fp))
    if(freadln(fp, str))
    {
      if(get_param(substr,str,"printer_path:",_PATH_LEN))
        strcpy(printer_path,substr);

      if(get_param(substr,str,"printer_list:",_PATH_LEN))
        strcpy(printer_list,substr);

      if(get_param(substr,str,"print_to:",_PATH_LEN))
        strcpy(print_to,substr);
    }

  fclose(fp);
  return(1);
}

int find_virtual_printer(char *prntr, char *v_prntr, char *send_to)
{
  FILE *fp;
  char str[_PATH_LEN + 1];
  char prntr_pfx[_PATH_LEN + 1];

  if(fp = fopen(printer_list,"r"))
  {
    while(!feof(fp))
      if(freadln(fp,str))
      {
        if(sscanf(str,"%s %s", prntr_pfx, v_prntr))
        {
          if(strncmp(prntr_pfx, prntr, strlen(prntr_pfx)) == 0)
          {
            strcpy(send_to, prntr + strlen(prntr_pfx));
            fclose(fp);
            return(1);
          }
        }
      }
    fclose(fp);
  }
  return 0;
}

int get_first_line(FILE *fp_in, int buffer[], char *str)
{
  int c, n;

  /* Get the first line of the page */
  n = 0;
  while (((c = fgetc( fp_in )) >= 0))
  {
    if( (char) c == '\f')
    {
      ungetc(c, fp_in );
      break;
    }

    buffer[n] = c;
    str[n] = (char) c;
    n++;

    if(n == _BUFFER_LEN)
      break;

    if( (char) c == '\n')
    {
      str[n-1] = '\0';
    break;
    }
  }

  return(n);
}

int get_printer_from_input(char *str, char *printer_name)
{
  *printer_name = '\0';

  if(print_to)
  {
    if(strncmp(str, print_to, strlen(print_to)) == 0)
    {
      /* Remove the new-line character */
      if(str[strlen(str)-1] == '\n')
        str[strlen(str)-1] = '\0';

      strcpy(printer_name, str + strlen(print_to));
      return 1;
    }
  }

return 0;
}

int print_file(FILE *fp_in, FILE *fp_out, int buffer[], int buffer_size, int printer_found)
{
  int c, n;

  if(!printer_found)
  {
    for(n = 0; n < buffer_size; n++)
    {
      c = buffer[n];
      fputc(c, fp_out);
    }
  }

  while ((c = fgetc( fp_in )) >= 0)
  {
    if(c != '\f')
      fputc(c, fp_out);
    else
      return 1;
  }

return 0;
}

int process_input(FILE *fp_in, char *cmd_line_printer, char *title, char *desc, int argc, char **argv)
{
  FILE *fp_out;
  int first_line[_BUFFER_LEN + 1];
  char first_line_chars[_BUFFER_LEN + 1];
  int first_line_len, printer_found;

  char printer_name[_PATH_LEN + 1];
  char new_printer_name[_PATH_LEN + 1];
  char v_printer[_PATH_LEN + 1];
  char send_to[_PATH_LEN + 1] = "";
  char str[_PATH_LEN + 1];
  int  c, n, rtn = 0;
  int more = 1, new_page;

  first_line_len = get_first_line(fp_in, first_line, first_line_chars);
  printer_found = get_printer_from_input(first_line_chars, printer_name);

  while(more)
  {
    new_page = 1;
    if(strcmp(cmd_line_printer, "-"))
    {
      strcpy(printer_name, cmd_line_printer);
      rtn = find_virtual_printer(printer_name, v_printer, send_to);
    }
    else
    {
      if(printer_found)
        rtn = find_virtual_printer(printer_name, v_printer, send_to);
      else
      {
        fprintf(stderr,"ERROR - No Printer Specified\n");
        exit(0);
      }
    }

    if(rtn)
    {
      /* Virtual Printer */
      sprintf(str,"%s %s \"%s\" \"%s\"", v_printer, send_to, title, desc);
    }
    else
    {
      /* "Real" Printer */
      sprintf(str,"%s", printer_path);
      for(n = 1; n < argc; n++)
      {
        strcat(str, " ");
        strcat(str, argv[n]);
      }               
    }

    if(fp_out = popen(str,"w"))
    {
      while(new_page)
      {
        new_page = print_file(fp_in, fp_out, first_line, first_line_len, printer_found);
        if(new_page)
        {
          first_line_len = get_first_line(fp_in, first_line, first_line_chars);
          printer_found = get_printer_from_input(first_line_chars, 
                           new_printer_name);

          if(printer_found && strcmp(printer_name, new_printer_name))
          {
            if(strcmp(cmd_line_printer, "-") == 0)
            {
              strcpy(printer_name, new_printer_name);
              new_page = 0;
            }
            else
              fputc('\f', fp_out);
          }
          else
            fputc('\f', fp_out);
        }     
        else
          more = 0;
      }
      pclose(fp_out);
    }
  }
  return 0;
}

int main( int argc, char **argv)
{
  FILE *fp_in;
  int opt;
  char printer[_PATH_LEN + 1] = "";
  char title[_PATH_LEN + 1] = "";
  char job_name[_PATH_LEN + 1] = "";

  printer[0] = 0;
  title[0] = 0;
  job_name[0] = 0;
 
  /* Do not display error message if invalid option found */
  opterr = 0;

  if(!load_conf())
    return -1;

  while((opt = getopt(argc,argv,"vhJ:P:T:")) != -1)
  {
    switch(opt)
    {
      case 'v':
      case 'h':  usage();

      case 'P':  if(!printer[0])
                   strncpy(printer, optarg, _PATH_LEN);
                 else
                 {
                   fprintf(stderr,"ERROR - Printer (%s) already specified\n", printer);
                   usage();
                 }
                 break;

      case 'J':  if(!job_name[0])
                     strncpy(job_name, optarg, _PATH_LEN);
                 else
                 {
                   fprintf(stderr,"ERROR - Duplicate parameter\n");
                   usage();;
                 }
                 break;

      case 'T':  if(!title[0])
                   strncpy(title, optarg, _PATH_LEN);
                 else
                 {
                   fprintf(stderr,"ERROR - Duplicate parameter\n");
                   usage();;
                 }
                 break;
    }
  }

  if(optind < argc)
  {
    for( ; optind < argc; optind++)
    {
      if((fp_in = fopen(argv[optind], "r")) == 0)
      {
        fprintf(stderr, "ERROR - Input file (%s) not found\n", argv[optind]);
        return -1;
      }
      process_input(fp_in, printer, title, job_name, argc, argv);
      fclose(fp_in);
    }
  }
  else
    process_input(stdin, printer, title, job_name, argc, argv);
  
  return 1;
}


/*    End of source file for FENPRINT    */
